@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-# Copyright 2021-2022 Mike Fährmann
+# Copyright 2021-2025 Mike Fährmann
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License version 2 as
@@ -9,7 +9,6 @@
 """Filesystem path handling"""
 
 import os
-import re
 import shutil
 import functools
 from . import util, formatter, exception
@@ -38,7 +37,7 @@ class PathFormat():
                 filename_fmt = extractor.filename_fmt
             elif isinstance(filename_fmt, dict):
                 self.filename_conditions = [
-                    (util.compile_expression(expr),
+                    (util.compile_filter(expr),
                      formatter.parse(fmt, kwdefault).format_map)
                     for expr, fmt in filename_fmt.items() if expr
                 ]
@@ -51,12 +50,13 @@ class PathFormat():
             raise exception.FilenameFormatError(exc)
 
         directory_fmt = config("directory")
+        self.directory_conditions = ()
         try:
             if directory_fmt is None:
                 directory_fmt = extractor.directory_fmt
             elif isinstance(directory_fmt, dict):
                 self.directory_conditions = [
-                    (util.compile_expression(expr), [
+                    (util.compile_filter(expr), [
                         formatter.parse(fmt, kwdefault).format_map
                         for fmt in fmts
                     ])
@@ -73,10 +73,15 @@ class PathFormat():
             raise exception.DirectoryFormatError(exc)
 
         self.kwdict = {}
-        self.directory = self.realdirectory = \
-            self.filename = self.extension = self.prefix = \
-            self.path = self.realpath = self.temppath = ""
         self.delete = False
+        self.prefix = ""
+        self.filename = ""
+        self.extension = ""
+        self.directory = ""
+        self.realdirectory = ""
+        self.path = ""
+        self.realpath = ""
+        self.temppath = ""
 
         extension_map = config("extension-map")
         if extension_map is None:
@@ -85,6 +90,7 @@ class PathFormat():
 
         restrict = config("path-restrict", "auto")
         replace = config("path-replace", "_")
+        conv = config("path-convert")
         if restrict == "auto":
             restrict = "\\\\|/<>:\"?*" if WINDOWS else "/"
         elif restrict == "unix":
@@ -93,10 +99,12 @@ class PathFormat():
             restrict = "\\\\|/<>:\"?*"
         elif restrict == "ascii":
             restrict = "^0-9A-Za-z_."
-        self.clean_segment = self._build_cleanfunc(restrict, replace)
+        elif restrict == "ascii+":
+            restrict = "^0-9@-[\\]-{ #-)+-.;=!}~"
+        self.clean_segment = _build_cleanfunc(restrict, replace, conv)
 
         remove = config("path-remove", "\x00-\x1f\x7f")
-        self.clean_path = self._build_cleanfunc(remove, "")
+        self.clean_path = _build_cleanfunc(remove, "")
 
         strip = config("path-strip", "auto")
         if strip == "auto":
@@ -115,7 +123,7 @@ class PathFormat():
             basedir = config("base-directory")
             sep = os.sep
             if basedir is None:
-                basedir = "." + sep + "gallery-dl" + sep
+                basedir = f".{sep}gallery-dl{sep}"
             elif basedir:
                 basedir = util.expand_path(basedir)
                 altsep = os.altsep
@@ -126,26 +134,14 @@ class PathFormat():
             basedir = self.clean_path(basedir)
         self.basedirectory = basedir
 
-    @staticmethod
-    def _build_cleanfunc(chars, repl):
-        if not chars:
-            return util.identity
-        elif isinstance(chars, dict):
-            def func(x, table=str.maketrans(chars)):
-                return x.translate(table)
-        elif len(chars) == 1:
-            def func(x, c=chars, r=repl):
-                return x.replace(c, r)
-        else:
-            return functools.partial(
-                re.compile("[" + chars + "]").sub, repl)
-        return func
-
     def open(self, mode="wb"):
         """Open file and return a corresponding file object"""
         try:
             return open(self.temppath, mode)
         except FileNotFoundError:
+            if "r" in mode:
+                # '.part' file no longer exists
+                return util.NullContext()
             os.makedirs(self.realdirectory)
             return open(self.temppath, mode)
 
@@ -155,8 +151,7 @@ class PathFormat():
             return self.check_file()
         return False
 
-    @staticmethod
-    def check_file():
+    def check_file(self):
         return True
 
     def _enum_file(self):
@@ -176,29 +171,30 @@ class PathFormat():
     def set_directory(self, kwdict):
         """Build directory path and create it if necessary"""
         self.kwdict = kwdict
-        sep = os.sep
 
-        segments = self.build_directory(kwdict)
-        if segments:
+        if segments := self.build_directory(kwdict):
             self.directory = directory = self.basedirectory + self.clean_path(
-                sep.join(segments) + sep)
+                os.sep.join(segments) + os.sep)
         else:
             self.directory = directory = self.basedirectory
 
         if WINDOWS and self.extended:
-            # Enable longer-than-260-character paths
-            directory = os.path.abspath(directory)
-            if directory.startswith("\\\\"):
-                directory = "\\\\?\\UNC\\" + directory[2:]
-            else:
-                directory = "\\\\?\\" + directory
-
-            # abspath() in Python 3.7+ removes trailing path separators (#402)
-            if directory[-1] != sep:
-                directory += sep
-
+            directory = self._extended_path(directory)
         self.realdirectory = directory
 
+    def _extended_path(self, path):
+        # Enable longer-than-260-character paths
+        path = os.path.abspath(path)
+        if not path.startswith("\\\\"):
+            path = "\\\\?\\" + path
+        elif not path.startswith("\\\\?\\"):
+            path = "\\\\?\\UNC\\" + path[2:]
+
+        # abspath() in Python 3.7+ removes trailing path separators (#402)
+        if path[-1] != os.sep:
+            return path + os.sep
+        return path
+
     def set_filename(self, kwdict):
         """Set general filename data"""
         self.kwdict = kwdict
@@ -224,6 +220,8 @@ class PathFormat():
                     self.temppath = self.realpath = self.realpath[:-1]
             elif not self.temppath:
                 self.build_path()
+        except exception.GalleryDLException:
+            raise
         except Exception:
             self.path = self.directory + "?"
             self.realpath = self.temppath = self.realdirectory + "?"
@@ -251,24 +249,22 @@ class PathFormat():
     def build_directory(self, kwdict):
         """Apply 'kwdict' to directory format strings"""
         segments = []
-        append = segments.append
         strip = self.strip
 
         try:
             for fmt in self.directory_formatters:
                 segment = fmt(kwdict).strip()
-                if strip:
+                if strip and segment not in {".", ".."}:
                     # remove trailing dots and spaces (#647)
                     segment = segment.rstrip(strip)
                 if segment:
-                    append(self.clean_segment(segment))
+                    segments.append(self.clean_segment(segment))
             return segments
         except Exception as exc:
             raise exception.DirectoryFormatError(exc)
 
     def build_directory_conditional(self, kwdict):
         segments = []
-        append = segments.append
         strip = self.strip
 
         try:
@@ -279,10 +275,10 @@ class PathFormat():
                 formatters = self.directory_formatters
             for fmt in formatters:
                 segment = fmt(kwdict).strip()
-                if strip:
+                if strip and segment != "..":
                     segment = segment.rstrip(strip)
                 if segment:
-                    append(self.clean_segment(segment))
+                    segments.append(self.clean_segment(segment))
             return segments
         except Exception as exc:
             raise exception.DirectoryFormatError(exc)
@@ -317,6 +313,11 @@ class PathFormat():
             pass
         return 0
 
+    def set_mtime(self, path=None):
+        if (mtime := (self.kwdict.get("_mtime_meta") or
+                      self.kwdict.get("_mtime_http"))):
+            util.set_mtime(self.realpath if path is None else path, mtime)
+
     def finalize(self):
         """Move tempfile to its target location"""
         if self.delete:
@@ -330,15 +331,72 @@ class PathFormat():
                 try:
                     os.replace(self.temppath, self.realpath)
                 except FileNotFoundError:
-                    # delayed directory creation
-                    os.makedirs(self.realdirectory)
+                    try:
+                        # delayed directory creation
+                        os.makedirs(self.realdirectory)
+                    except FileExistsError:
+                        # file at self.temppath does not exist
+                        return False
                     continue
                 except OSError:
                     # move across different filesystems
-                    shutil.copyfile(self.temppath, self.realpath)
+                    try:
+                        shutil.copyfile(self.temppath, self.realpath)
+                    except FileNotFoundError:
+                        try:
+                            os.makedirs(self.realdirectory)
+                        except FileExistsError:
+                            return False
+                        shutil.copyfile(self.temppath, self.realpath)
                     os.unlink(self.temppath)
                 break
 
-        mtime = self.kwdict.get("_mtime")
-        if mtime:
-            util.set_mtime(self.realpath, mtime)
+        self.set_mtime()
+
+
+def _build_convertfunc(func, conv):
+    if len(conv) <= 1:
+        conv = formatter._CONVERSIONS[conv]
+        return lambda x: conv(func(x))
+
+    def convert_many(x):
+        x = func(x)
+        for conv in convs:
+            x = conv(x)
+        return x
+    convs = [formatter._CONVERSIONS[c] for c in conv]
+    return convert_many
+
+
+def _build_cleanfunc(chars, repl, conv=None):
+    if not chars:
+        func = util.identity
+    elif isinstance(chars, dict):
+        if 0 not in chars:
+            chars = _process_repl_dict(chars)
+            chars[0] = None
+
+        def func(x):
+            return x.translate(table)
+        table = str.maketrans(chars)
+    elif len(chars) == 1:
+        def func(x):
+            return x.replace(chars, repl)
+    else:
+        func = functools.partial(util.re(f"[{chars}]").sub, repl)
+    return _build_convertfunc(func, conv) if conv else func
+
+
+def _process_repl_dict(chars):
+    # can't modify 'chars' while *directly* iterating over its keys
+    for char in [c for c in chars if len(c) > 1]:
+        if len(char) == 3 and char[1] == "-":
+            citer = range(ord(char[0]), ord(char[2])+1)
+        else:
+            citer = char
+
+        repl = chars.pop(char)
+        for c in citer:
+            chars[c] = repl
+
+    return chars
